// Copyright (C) Edward W. Felten, 2003.
// Recent changes, A. Zubow

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import java.io.IOException;

public class ChatServer {
	public static final int portNum = Config.getAsInt("ServerPortNum");

	// activeSenders is the list of clients that are currently active.
	private Set activeSenders = Collections.synchronizedSet(new HashSet());

	public ChatServer() {
		// This constructor never returns, unless there is an error.
		try {
			ServerSocket ss;
			ss = new ServerSocket(portNum);
			for (;;) {
				// wait for a new client to connect, then hook it up properly
				Socket sock = ss.accept();
				InputStream in = sock.getInputStream();
				OutputStream out = sock.getOutputStream();
				SenderThread sender = new SenderThread(out);
				new ReceiverThread(in, sender);
			}
		} catch (IOException x) {
			System.err.println("Dying: IOException: " + x.getMessage());
		}
	}
	
	// entry point
	public static void main(String[] argv) {
		new ChatServer();
	}

	class SenderThread extends Thread {
		// forwards messages to a client
		// messages are queued when somebody calls queueForSending
		// we take them from the queue and send them along

		private OutputStream out;

		private Queue queue;

		SenderThread(OutputStream outStream) {
			out = outStream;
			queue = new Queue();
			activeSenders.add(this);
			start();
		}

		public void queueForSending(byte[] message) {
			// Queue a message, to be sent as soon as possible.
			// We queue messages, rather than sending them immediately, because
			// sending immediately would cause us to block, if the client
			// had fallen behind in processing his incoming messages. If we
			// blocked, the processing of incoming messages would be frozen,
			// which would be Very Bad. By queueing messages, we can ensure that
			// the processing of incoming messages never stalls, no matter how
			// badly clients behave.

			queue.put(message);
		}

		public void run() {
			// suck messages out of the queue and send them out
			try {
				for (;;) {
					Object o = queue.get();
					byte[] barr = (byte[]) o;
					out.write(barr);
					out.flush();
				}
			} catch (IOException x) {
				// unexpected exception -- stop relaying messages
				x.printStackTrace();
				try {
					out.close();
				} catch (IOException x2) { x2.printStackTrace(); }
			}
			activeSenders.remove(this);
		}
	}

	class ReceiverThread extends Thread {
		// receives messages from a client, and forwards them to everybody else

		private InputStream in;

		private SenderThread me;

		private String userName;

		private byte[] userNameBytes = null;

		ReceiverThread(InputStream inStream, SenderThread mySenderThread) {
			in = inStream;
			me = mySenderThread;
			start();
		}

		public void run() {
			// get first line, which is the client's username
			ByteArrayOutputStream baos;
			
			try {
				baos = getOneLine(false);
			} catch (CustomIOException cx) { // I/O exception -- display what we have 
				baos = cx.getLastInput();
			}
			String name = new String(baos.toByteArray());
			name = name.substring(0, name.length() - 1); // trim trailing
			// carriage return
			userName = name;
			name = "[" + name + "] ";
			userNameBytes = name.getBytes();

			// get subsequent lines, and sent them to the other clients
			for (;;) {
				try {
					baos = getOneLine(true);
				} catch (CustomIOException cx) {
					sendToOthers(cx.getLastInput()); // send what we have
					activeSenders.remove(me); // unregister associated client
					return;					
				}
				sendToOthers(baos);
			}
		}

		private ByteArrayOutputStream getOneLine(boolean prependUserName)
			throws CustomIOException {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			// read in a message, terminated by carriage-return
			// buffer the message in baos, until we see EOF or carriage-return
			// return the message we got
			// if prependUserName is true, then prepend the username of the
			// sender to the message before we return it
			try {
				if (prependUserName)
					baos.write(userNameBytes);
				int c;
				do {
					c = in.read();
					if (c == -1) {
						// got EOF -- return what we have, then quit
						throw new CustomIOException(baos);
					}
					baos.write(c);
				} while (c != '\n');
				// return what we have -- note: this includes a final CR
				return (baos);
			} catch (IOException x) {
				// return what we have, then quit
				CustomIOException ex = new CustomIOException(baos);
				ex.initCause(x);
				throw ex;
			}
		}

		// stArr is a dummy variable, used to make toArray happy below
		private final SenderThread[] stArr = new SenderThread[1];

		private void sendToOthers(ByteArrayOutputStream baos) {
			// extract the contents of baos, and queue them for sending to all
			// other clients (but not to ourself);
			// also, reset baos so it is empty and can be reused

			byte[] message = baos.toByteArray();
			baos.reset();

			SenderThread[] guys = (SenderThread[]) (activeSenders
					.toArray(stArr));
			for (int i = 0; i < guys.length; ++i) {
				SenderThread st = guys[i];
				if (st != me)
					st.queueForSending(message);
			}
		}
	}
	
	/* Custom exception to handle i/o errors. */
	class CustomIOException extends IOException {
		
		static final long serialVersionUID = 5186231825499166944L;
	
		ByteArrayOutputStream lastInput;
		
		public CustomIOException(ByteArrayOutputStream lastInput) {
			super();
			this.lastInput = lastInput;
		}
		
		public ByteArrayOutputStream getLastInput() {
			return lastInput;
		}
	}
}
